#ifndef CMAYBE_H
#define CMAYBE_H

#define MAYBE_TYPE(name, type) struct maybe_##name {type value; char is_value;}
#define MAYBE(name) struct maybe_##name

#define ENABLE_RETURN(name) MAYBE(name) _return_value
#define RETURN_VALUE(x) {\
	_return_value.is_value = 1;\
	_return_value.value = x;\
	return _return_value;\
}
#define RETURN_NOTHING() {\
	_return_value.is_value = 0;\
	return _return_value;\
}

#define IS_VALUE(x) if ((x).is_value)
#define IS_NOT_VALUE(x) if (!(x).is_value)
#define VALUE(x) (x).value

struct _free_list {
	struct _free_list *next;
	void *ptr;
	void (*freer)(void*);
};
#define ENABLE_TRY() \
	struct _free_list *_allocations = NULL, *_free_list_node;\
	void *_allocation;\
	jmp_buf _try_fail_jmp_buf;\
	if (setjmp(_try_fail_jmp_buf)) {\
		TRY_FAIL_HANDLE();\
		RETURN_NOTHING();\
	}
#define TRY_TYPE(name) MAYBE(name) _try_tmp_##name
#define TRY(name, maybe_value) (\
	_try_tmp_##name = maybe_value,\
	(_try_tmp_##name.is_value ? 0 :\
		longjmp(_try_fail_jmp_buf, 1)\
	),\
	_try_tmp_##name.value\
)

#define TRY_HOF_HANDLE(freer_func, pointer) (\
	_allocation = pointer,\
	_free_list_node = malloc(sizeof(struct _free_list)),\
	(_free_list_node == NULL ? (\
		free(_allocation),\
		longjmp(_try_fail_jmp_buf, 1)\
	) : (\
		_free_list_node->next = _allocations,\
		_free_list_node->ptr = _allocation,\
		_free_list_node->freer = freer_func,\
		_allocations = _free_list_node\
	)),\
	_allocation\
)
#define FREE_ON_TRY_FAIL(pointer) TRY_HOF_HANDLE(free, pointer)

// free() is passed as an argument to avoid errors in case the program uses
// no TRY functionality and doesn't include stdlib.h
static void _remove_free_list(struct _free_list **head, void *ptr, void (*free)(void*)) {
	struct _free_list **current = head;
	while (*current != NULL) {
		if (current[0]->ptr == ptr) {
			struct _free_list *deleted = *current;
			*current = deleted->next;
			free(deleted);
			break;
		}
		current = &current[0]->next;
	}
}

#define TRY_HOF_REMOVE(allocation) _remove_free_list(&_allocations, allocation, free)
#define TRY_FAIL_HANDLE() \
	while (_allocations != NULL) {\
		struct _free_list *current = _allocations;\
		void (*freer)(void*) = current->freer;\
		freer(current->ptr);\
		_allocations = current->next;\
		free(current);\
	}

static void _replace_free_list(struct _free_list **head, void *from, void *to) {
	struct _free_list **current = head;
	while (*current != NULL) {
		if (current[0]->ptr == from) {
			current[0]->ptr = to;
			break;
		}
		current = &current[0]->next;
	}
}
#define TRY_HOF_REPLACE(old_ptr, new_ptr) _replace_free_list(&_allocations, old_ptr, new_ptr)

#endif
